package hr.pmf.aklobucar.extended.scenariosolver;


import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.*;
import java.util.stream.Collectors;

public class ScenarioSolver {

    public static String PARAMETERS = "parameters.json";
    public static String RESOURCES = "resources";
    public static String LOG_LEVEL = "INFO";

    public static void createFileIfNotExisting(String folder, String file) throws IOException {
        if (!Files.exists(Paths.get(RESOURCES, folder))) {
            Files.createDirectory(Paths.get(RESOURCES, folder));
        }
        if (!Files.exists(Paths.get(RESOURCES, folder, file))) {
            Files.createFile(Paths.get(RESOURCES, folder, file));
        }
    }

    public static void resetFile(String folder, String file) throws IOException {
        Files.write(Paths.get(RESOURCES, folder, file), ScenarioSolution.getHeaders().getBytes());
    }

    public static void log(String input, String currentLevel) {
        if (LOG_LEVEL.equals("DEBUG")) {
            System.out.println(input);
        } else if (currentLevel.equals(LOG_LEVEL)) {
            System.out.println(input);
        }

    }


    public static void main(String[] args) throws IOException {

        Gson gson = new GsonBuilder().setPrettyPrinting().create();
        InputParameters inputParameters = gson.fromJson(new String(Files.readAllBytes(Paths.get(RESOURCES, PARAMETERS)), StandardCharsets.UTF_8), InputParameters.class);
        LOG_LEVEL = inputParameters.getLogLevel();

        String INPUT_FOLDER = "input_interval_" + inputParameters.getNodes();
        String OUTPUT_FOLDER = "output_scenario_solver";
        String OUTPUT_FILE = "results_" + inputParameters.getNodes() + "_nodes_" + (inputParameters.getType().equals("LOCAL_SEARCH") ? (inputParameters.getAreaOrIterations() + "_area") : (inputParameters.getAreaOrIterations()) + "_iterations") + ".csv";


        createFileIfNotExisting(OUTPUT_FOLDER, OUTPUT_FILE);
        if (!inputParameters.append) {
            resetFile(OUTPUT_FOLDER, OUTPUT_FILE);
        }


        List<ScenarioParameter> scenarioParameters = new ArrayList<>();
        for (double factor = inputParameters.factorStart; factor <= inputParameters.factorEnd; factor = round(inputParameters.factorStep + factor)) {
            for (double chance = inputParameters.chanceStart; chance <= inputParameters.chanceEnd; chance = round(inputParameters.chanceStep + chance)) {
                scenarioParameters.add(new ScenarioParameter(factor, chance));
            }
        }

        scenarioParameters.forEach(scenarioParameter -> {

            HashMap<String, List<ScenarioSolution>> solutions = new HashMap<>();

            File[] files = Paths.get(RESOURCES, INPUT_FOLDER).toFile().listFiles();

            assert files != null;
            for (File file : files) {
                if (file.isDirectory()) {
                    continue;
                }
                String scenario = file.getName().split("_interval")[0];
                String index = file.getName().split("interval_")[1].split(".txt")[0];

                solutions.computeIfAbsent(scenario, k -> new ArrayList<>());

                int children = Integer.parseInt(scenario.split("_")[2]);
                int maxWeight = Integer.parseInt(scenario.split("_")[3]);
                List<String> lines = null;
                try {
                    lines = Files.readAllLines(file.toPath()).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());
                } catch (IOException e) {
                    e.printStackTrace();
                }
                String infoLine = lines.get(0);
                String[] parts = infoLine.split("\\s+");
                int verticesNum = Integer.parseInt(parts[0]);
                int edgesNum = Integer.parseInt(parts[1]);

                List<Node> tree = new ArrayList<>();
                for (int verticesIndex = 0; verticesIndex < verticesNum; verticesIndex++) {
                    tree.add(new Node(verticesIndex));
                }

                lines.subList(1, edgesNum + 1).forEach(edgeLine -> {
                    String[] edgeParts = edgeLine.split("\\s+");
                    int parent = Integer.parseInt(edgeParts[1]);
                    int child = Integer.parseInt(edgeParts[2]);

                    tree.get(parent).children.add(tree.get(child));
                    tree.get(child).parent = tree.get(parent);
                });

                String intervalLine = lines.get(lines.size() - 1);
                String[] intervals = intervalLine.split("\\s+");
                for (int intervalIndex = 0; intervalIndex < intervals.length; intervalIndex++) {
                    String interval = intervals[intervalIndex];
                    Node node = tree.get(intervalIndex);
                    node.lower = Double.parseDouble(interval.split(",")[0]);
                    node.upper = Double.parseDouble(interval.split(",")[1]);
                    node.weight = (node.upper + node.lower) / 2.0;
                }

                log("Running scenario " + scenario + " on area/iterations: " + inputParameters.getAreaOrIterations(), "INFO");

                List<Double> currentBestScenario = tree.stream().map(node -> node.weight).collect(Collectors.toList());
                long start = System.nanoTime();
                double bestRegret = calculateRegret(tree, currentBestScenario);
                long initialRegretTime = System.nanoTime() - start;
                double initialRegret = bestRegret;

                log("Initial regret " + initialRegret, "INFO");

                List<Long> localTimes = new ArrayList<>();
                if (inputParameters.getType().equals("LOCAL_SEARCH")) {
                    while (true) {
                        log("Running local search", "DEBUG");
                        long startLocalSearch = System.nanoTime();
                        boolean improved = false;
                        List<List<Double>> localScenarios = new ArrayList<>();

                        for (int i = 0; i < inputParameters.getAreaOrIterations(); i++) {
                            localScenarios.add(mutate(tree, currentBestScenario, scenarioParameter.mutationFactor, scenarioParameter.mutationChance));
                        }
                        for (List<Double> localScenario : localScenarios) {
                            double regret = calculateRegret(tree, localScenario);
                            if (regret < bestRegret) {
                                currentBestScenario = localScenario;
                                bestRegret = regret;
                                improved = true;
                                log("Found better regret in local search: " + regret, "DEBUG");
                            }
                        }
                        localTimes.add(System.nanoTime() - startLocalSearch);
                        if (!improved) {
                            log("Better regret not found. Cancelling local search", "DEBUG");
                            break;
                        }
                    }
                } else {
                    int counter = 0;
                    while (counter < inputParameters.getAreaOrIterations()) {
                        long startIteration = System.nanoTime();
                        counter += 1;

                        List<Double> mutatedScenario = mutate(tree, currentBestScenario, scenarioParameter.mutationFactor, scenarioParameter.mutationChance);

                        double regret = calculateRegret(tree, mutatedScenario);

                        if (regret < bestRegret) {
                            currentBestScenario = mutatedScenario;
                            bestRegret = regret;
                        }
                        localTimes.add(System.nanoTime() - startIteration);
                    }
                }

                long totalTime = System.nanoTime() - start;

                double finalRegret = bestRegret;

                log("Best regeret: " + bestRegret, "INFO");
                double improvement = round((1 - (bestRegret / initialRegret)) * 100);
                log("Improvement: " + improvement, "INFO");

                double averageLocalSearchTime = round(localTimes.stream().mapToLong(Long::longValue).sum() / (localTimes.size() * 1.0));

                solutions.get(scenario).add(new ScenarioSolution(
                        index,
                        inputParameters.getNodes(),
                        children,
                        maxWeight,
                        scenarioParameter.mutationChance,
                        scenarioParameter.mutationFactor,
                        inputParameters.getAreaOrIterations(),
                        initialRegretTime,
                        initialRegret,
                        finalRegret,
                        improvement,
                        averageLocalSearchTime,
                        totalTime
                ));
            }

            solutions.keySet().forEach(scenario -> {

                int children = Integer.parseInt(scenario.split("_")[2]);
                int maxWeight = Integer.parseInt(scenario.split("_")[3]);

                double size = solutions.get(scenario).size() * 1.0;

                Double averageImprovement = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getImprovement).sum() / size);
                Double averageInitialRegret = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getInitialRegret).sum() / size);
                Double averageFinalRegret = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getFinalRegret).sum() / size);


                Double averageInitialRegretTime = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getInitialRegretTime).sum() / (size * 1000));
                Double averageLocalSearchTime = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getAverageLocalSearchTime).sum() / (size * 1000));
                Double averageTotalTime = round(solutions.get(scenario).stream().mapToDouble(ScenarioSolution::getTotalTime).sum() / (size * 1000));

                Double averageSlowdown = round(solutions.get(scenario).stream()
                        .mapToDouble(solution -> solution.getTotalTime() / (solution.getInitialRegretTime() * 1.0))
                        .sum() / size);

                AverageScenarioSolution averageScenarioSolution = new AverageScenarioSolution(
                        inputParameters.getNodes(), children, maxWeight, scenarioParameter.mutationChance, scenarioParameter.mutationFactor,
                        inputParameters.getAreaOrIterations(),
                        averageInitialRegret, averageInitialRegretTime, averageFinalRegret,
                        averageImprovement, averageLocalSearchTime, averageTotalTime, averageSlowdown);

                solutions.get(scenario).forEach(scenarioSolution -> {
                    try {
                        Files.write(Paths.get(RESOURCES, OUTPUT_FOLDER, OUTPUT_FILE), scenarioSolution.getCsv().getBytes(), StandardOpenOption.APPEND);
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                });


            });
        });

    }

    public static Double calculateRegret(List<Node> tree, List<Double> scenario) {
        Solution standardSolution = solve(tree, scenario);

        List<Double> worstScenario = tree.stream().map(node -> {
            if (standardSolution.vertices.contains(node.index)) {
                return node.lower;
            } else {
                return node.upper;
            }
        }).collect(Collectors.toList());

        double standardSolutionWorstScenarioWeight = standardSolution.vertices.stream().map(worstScenario::get).reduce(0.0, Double::sum);

        Solution worstScenarioSolution = solve(tree, worstScenario);

        return worstScenarioSolution.weight - standardSolutionWorstScenarioWeight;
    }

    public static List<Double> mutate(List<Node> tree, List<Double> currentScenario, double mutationFactor, double mutationChance) {
        List<Double> mutatedList = new ArrayList<>();

        tree.forEach(node -> {
            double scenarioWeight = currentScenario.get(node.index);
            double interval = node.upper - node.lower;
            double mutation = interval * (mutationFactor * Math.random());
            if (Math.random() > mutationChance) {
                mutatedList.add(scenarioWeight);
            } else {
                if (Math.random() > 0.5) {
                    mutatedList.add(Math.max(scenarioWeight - mutation, node.lower));
                } else {
                    mutatedList.add(Math.min(scenarioWeight + mutation, node.upper));
                }

            }
        });

        return mutatedList;
    }

    public static Solution solve(List<Node> tree, List<Double> scenario) {
        for (int verticesIndex = (tree.size() - 1); verticesIndex >= 0; verticesIndex--) {

            Node currentNode = tree.get(verticesIndex);

            List<Node> children = currentNode.children;

            currentNode.setIncluded.add(currentNode);
            for (Node value : children) {
                currentNode.setIncluded.addAll(value.setExcluded);
            }

            List<Node> bestCombinationSet = new ArrayList<>();
            for (Node child : children) {
                if (child.setExcluded.stream().map(node -> scenario.get(node.index)).reduce(0.0, Double::sum) > child.setIncluded.stream().map(node -> scenario.get(node.index)).reduce(0.0, Double::sum)) {
                    bestCombinationSet.addAll(child.setExcluded);
                } else {
                    bestCombinationSet.addAll(child.setIncluded);
                }
                child.setExcluded.clear();
                ;
                child.setIncluded.clear();
            }
            currentNode.setExcluded = bestCombinationSet;
        }
        Node root = tree.get(0);

        double weightIncluded = root.setIncluded.stream().map(node -> scenario.get(node.index)).reduce(0.0, Double::sum);
        double weightExcluded = root.setExcluded.stream().map(node -> scenario.get(node.index)).reduce(0.0, Double::sum);
        Solution solution;
        if (weightExcluded > weightIncluded) {
            solution = new Solution(root.setExcluded.stream().map(node -> node.index).collect(Collectors.toList()), weightExcluded);
        } else {
            solution = new Solution(root.setIncluded.stream().map(node -> node.index).collect(Collectors.toList()), weightIncluded);
        }
        root.setIncluded.clear();
        root.setExcluded.clear();
        return solution;
    }

    public static Double round(Double num) {
        return Math.round(num * 1000.0) / 1000.0;
    }
}
